@@ -10,24 +10,29 @@ from copy import deepcopy
 from random import random
 
 import requests
-from urllib3.connection import ConnectionError as Urllib3Error
-from requests.exceptions import ConnectionError as RequestsError
+from urllib3.exceptions import HTTPError, PoolError
+from requests.exceptions import RequestException
 
-if sys.version_info[0] == 3:
+if sys.version_info[0] >= 3:
     string_types = str
 else:
     string_types = basestring
+    warnings.warn('Module pdpyras will no longer support Python 2.7 as of '
+        'June 21, 2021.')
 
-__version__ = '4.1'
-
+__version__ = '4.1.3'
 
 # These are API resource endpoints/methods for which multi-update is supported
-valid_multi_update_paths = [
+VALID_MULTI_UPDATE_PATHS = [
     ('incidents', '{index}'),
     ('incidents', '{id}', 'alerts', '{index}'),
     ('priorities', '{index}'),
 ]
 
+ITERATION_LIMIT = 1e4
+
+TIMEOUT = 5
+
 #########################
 ### UTILITY FUNCTIONS ###
 #########################
@@ -114,14 +119,14 @@ def resource_envelope(method):
         but with "r" prepended.
     :returns: A callable object; the reformed method
     """
-    global valid_multi_update_paths
+    global VALID_MULTI_UPDATE_PATHS
     http_method = method.__name__.lstrip('r')
     def call(self, path, **kw):
         pass_kw = deepcopy(kw) # Make a copy for modification
         nodes = tokenize_url_path(path, baseurl=self.url)
         is_index = nodes[-1] == '{index}'
         resource = nodes[-2]
-        multi_put = http_method == 'put' and nodes in valid_multi_update_paths
+        multi_put = http_method == 'put' and nodes in VALID_MULTI_UPDATE_PATHS
         envelope_name_single = object_type(resource) # Usually the "type"
         if is_index and http_method=='get' or multi_put:
             # Plural resource name, for index action (GET /<resource>), or for
@@ -207,7 +212,7 @@ def tokenize_url_path(url, baseurl='https://api.pagerduty.com'):
     :type method: str
     :type url: str
     :type baseurl: str
-    :rtype: tuple 
+    :rtype: tuple
     """
     urlnparams = url.split('#')[0].split('?') # Ignore all #'s / params
     url_nodes = urlnparams[0].lstrip('/').split('/')
@@ -232,7 +237,7 @@ def tokenize_url_path(url, baseurl='https://api.pagerduty.com'):
     # Tokenize / classify the URL now:
     tokenized_nodes = [path_nodes[0]]
     if len(path_nodes) >= 3:
-        # It's an endpoint like one of the following 
+        # It's an endpoint like one of the following
         # /{resource}/{id}/{sub-resource}
         # We're interested in {resource} and {sub_resource}.
         # More deeply-nested endpoints not known to exist.
@@ -324,16 +329,19 @@ class PDSession(requests.Session):
     upon receiving a HTTP error.
     """
 
-    retry = None
+    retry = {}
     """
     A dict defining the retry behavior for each HTTP response status code.
 
+    Note, any value set for this class variable will not be reflected in
+    instances and so it must be set separately for each instance.
+
     Each key in this dictionary represents a HTTP response code. The behavior is
     specified by the value at each key as follows:
 
     * ``-1`` to retry infinitely
     * ``0`` to return the `requests.Response`_ object and exit (which is the
-      default behavior) 
+      default behavior)
     * ``n``, where ``n > 0``, to retry ``n`` times (or up
       to :attr:`max_http_attempts` total for all statuses, whichever is
       encountered first), and raise a :class:`PDClientError` after that many
@@ -490,7 +498,7 @@ class PDSession(requests.Session):
         # Merge, but do not replace, any headers specified in keyword arguments:
         if 'headers' in kwargs:
             my_headers.update(kwargs['headers'])
-        req_kw.update({'headers': my_headers, 'stream': False})
+        req_kw.update({'headers': my_headers, 'stream': False, 'timeout': TIMEOUT})
         # Compose/normalize URL whether or not path is already a complete URL
         if url.startswith(self.url) or not self.url:
             my_url = url
@@ -501,15 +509,15 @@ class PDSession(requests.Session):
             try:
                 response = self.parent.request(method, my_url, **req_kw)
                 self.postprocess(response)
-            except (Urllib3Error, RequestsError) as e:
+            except (HTTPError, PoolError, RequestException) as e:
                 network_attempts += 1
                 if network_attempts > self.max_network_attempts:
                     raise PDClientError("Non-transient network error; exceeded "
                         "maximum number of attempts (%d) to connect to the "
                         "API"%self.max_network_attempts)
                 sleep_timer *= self.cooldown_factor()
-                self.log.debug("Connection error: %s; retrying in %g seconds.",
-                    e, sleep_timer)
+                self.log.debug("HTTP or network error: %s: %s; retrying in %g "
+                    "seconds.", e.__class__.__name__, e, sleep_timer)
                 time.sleep(sleep_timer)
                 continue
 
@@ -602,7 +610,7 @@ class PDSession(requests.Session):
     def stagger_cooldown(self, val):
         if type(val) not in [float, int] or val<0:
             raise ValueError("Cooldown randomization factor stagger_cooldown "
-                "must be a positive real number") 
+                "must be a positive real number")
         self._stagger_cooldown = val
 
     @property
@@ -635,7 +643,7 @@ class EventsAPISession(PDSession):
 
     @property
     def auth_header(self):
-        return {'X-Routing-Key': self.api_key}
+        return {}
 
     def acknowledge(self, dedup_key):
         """
@@ -684,10 +692,16 @@ class EventsAPISession(PDSession):
         :returns:
             The deduplication key of the incident, if any.
         """
+
         actions = ('trigger', 'acknowledge', 'resolve')
         if action not in actions:
             raise ValueError("Event action must be one of: "+', '.join(actions))
-        event = {'event_action':action}
+
+        event = {
+            'event_action':action,
+            'routing_key': self.api_key
+        }
+
         event.update(properties)
         if isinstance(dedup_key, string_types):
             event['dedup_key'] = dedup_key
@@ -790,7 +804,7 @@ class APISession(PDSession):
     :members:
     """
 
-    api_call_counts = None 
+    api_call_counts = None
     """A dict object recording the number of API calls per endpoint"""
 
     api_time = None
@@ -827,7 +841,7 @@ class APISession(PDSession):
     @property
     def auth_type(self):
         """
-        Defines the method of API authentication. 
+        Defines the method of API authentication.
 
         By default this is "token"; if "oauth2", the API key will be used.
         """
@@ -910,12 +924,12 @@ class APISession(PDSession):
         query_params.update({'query':query})
         # When determining uniqueness, web/the API doesn't care about case.
         simplify = lambda s: s.lower()
-        search_term = simplify(query) 
+        search_term = simplify(query)
         equiv = lambda s: simplify(s[attribute]) == search_term
         obj_iter = self.iter_all(resource, params=query_params)
         return next(iter(filter(equiv, obj_iter)), None)
 
-    def iter_all(self, path, params=None, paginate=True, page_size=None, 
+    def iter_all(self, path, params=None, paginate=True, page_size=None,
             item_hook=None, total=False):
         """
         Iterator for the contents of an index endpoint or query.
@@ -940,8 +954,8 @@ class APISession(PDSession):
             pagination yet, i.e. "nested" endpoints like (as of this writing):
             ``/users/{id}/contact_methods`` and ``/services/{id}/integrations``
         :param page_size:
-            If set, the ``page_size`` argument will override the ``default_page_size`` 
-            parameter on the session and set the ``limit`` parameter to a custom 
+            If set, the ``page_size`` argument will override the ``default_page_size``
+            parameter on the session and set the ``limit`` parameter to a custom
             value (default is 100), altering the number of pagination results.
         :param item_hook:
             Callable object that will be invoked for each iteration, i.e. for
@@ -984,10 +998,21 @@ class APISession(PDSession):
 
         more = True
         offset = 0
+        if params is not None:
+            offset = int(params.get('offset', 0))
         n = 0
         while more: # Paginate through all results
             if paginate:
                 data['offset'] = offset
+                highest_record_index = int(data['offset']) + int(data['limit'])
+                if highest_record_index > ITERATION_LIMIT:
+                    self.log.warn("Stopping iteration on endpoint \"%s\" at "
+                        "limit+offset=%d as this exceeds the maximum permitted "
+                        "by the API (%d). The retrieved data may be incomplete."
+                        "For more information, see: %s", path,
+                        highest_record_index, ITERATION_LIMIT,
+                        'https://developer.pagerduty.com/docs/rest-api-v2/pagination')
+                    break
             r = self.get(path, params=data.copy())
             if not r.ok:
                 if self.raise_if_http_error:
@@ -999,8 +1024,8 @@ class APISession(PDSession):
                 break
             try:
                 response = r.json()
-            except ValueError: 
-                self.log.debug("Stopping iteration on endpoint %s; API "
+            except ValueError:
+                self.log.debug("Stopping iteration on endpoint \"%s\"; API "
                     "responded with invalid JSON.", path)
                 break
             if 'limit' in response:
@@ -1011,13 +1036,13 @@ class APISession(PDSession):
                 if 'more' in response:
                     more = response['more']
                 else:
-                    self.log.debug("Pagination is enabled in iteration, but the" 
+                    self.log.debug("Pagination is enabled in iteration, but the"
                         " index endpoint %s responded with no \"more\" property"
                         " in the response. Only the first page of results, "
                         "however many can be gotten, will be included.", path)
                 if 'total' in response:
                     total_count = response['total']
-                else: 
+                else:
                     self.log.debug("Pagination and the \"total\" parameter "
                         "are enabled in iteration, but the index endpoint %s "
                         "responded with no \"total\" property in the response. "
@@ -1025,7 +1050,7 @@ class APISession(PDSession):
                         "first retrieving all records.", path)
                 offset += data['limit']
             for result in response[r_name]:
-                n += 1 
+                n += 1
                 # Call a callable object for each item, i.e. to print progress:
                 if hasattr(item_hook, '__call__'):
                     item_hook(result, n, total_count)
@@ -1036,7 +1061,7 @@ class APISession(PDSession):
         """
         Performs a GET request, returning the JSON-decoded body as a dictionary
 
-        :raises PDClientError: In the event of HTTP error 
+        :raises PDClientError: In the event of HTTP error
         """
         return self.get(path, **kw)
 
@@ -1045,7 +1070,7 @@ class APISession(PDSession):
         """
         Performs a POST request, returning the JSON-decoded body as a dictionary
 
-        :raises PDClientError: In the event of HTTP error 
+        :raises PDClientError: In the event of HTTP error
         """
         return self.post(path, **kw)
 
@@ -1112,9 +1137,9 @@ class APISession(PDSession):
 
         :param method:
             Method of the request
-        :param response: 
+        :param response:
             Response object
-        :param suffix: 
+        :param suffix:
             Optional suffix to append to the key
         :type method: str
         :type response: `requests.Response`_
@@ -1138,7 +1163,6 @@ class APISession(PDSession):
             "#x_request_id=%s|#date=%s|#wall_time_s=%g", method, url, status,
             request_id, request_date, request_time)
         if int(status/100) == 5:
-            request_id = response.headers['x-request-id']
             self.log.error("PagerDuty API server error (%d)! "
                 "For additional diagnostics, contact PagerDuty support "
                 "and reference x_request_id=%s / date=%s",
@@ -1279,7 +1303,7 @@ class APISession(PDSession):
         """Truncated token for secure display/identification purposes."""
         return last_4(self.api_key)
 
-class PDClientError(Exception): 
+class PDClientError(Exception):
     """
     General API errors base class.
     """
